#include <os.h>

#include "level.h"

#include "enemies.h"
#include "utils.h"


#define STARTDELAY_BLINKY 0
#define STARTDELAY_PINKY 1
#define STARTDELAY_INKEY 2
#define STARTDELAY_CLYDE 5

static uint16_t enemyx[4];
static uint16_t enemyy[4];
static uint8_t enemydir[4];
static uint8_t enemynextdir[4];
static uint8_t enemymode[4];
static uint16_t enemyvel[4] = {192,180,184,170};

inline uint16_t* getEnemyx(){
    return enemyx;
}

inline uint16_t* getEnemyy(){
    return enemyy;
}

inline uint8_t* getEnemymode(){
    return enemymode;
}

inline void reversedirs(){
    int i;
    for(i=0;i<4;i++){
        if(enemynextdir[i] == DIR_LEFT)
            enemynextdir[i] = DIR_RIGHT;
        else if(enemydir[i] == DIR_RIGHT)
            enemynextdir[i] = DIR_LEFT;
        else if(enemydir[i] == DIR_UP)
            enemynextdir[i] = DIR_DOWN;
        else if(enemydir[i] == DIR_DOWN)
            enemynextdir[i] = DIR_UP;
    }
}

inline void startFrightmode(){
    setuint8array(enemymode,2,2,2,2);
    reversedirs();
}

inline void endFrightmode(){
    setuint8array(enemymode,0,0,0,0);
}

inline void resetEnemymode(uint8_t enemynum){
    enemymode[enemynum] = 0;
}

inline void resetEnemy(uint8_t enemynum){
    enemyx[enemynum] = (14*7)<<8;
    enemyy[enemynum] = (11*7+3)<<8;
    enemydir[enemynum] = DIR_LEFT;
    enemynextdir[enemynum] = DIR_LEFT;
}

static inline uint8_t distance(uint16_t x1, uint16_t y1, uint16_t x2, uint16_t y2){
    return abs(x1-x2) + abs(y1-y2);//here it isn't necessary to calculate the geometric distance
}

uint8_t newdir(uint8_t enemydir, uint16_t enemyx, uint16_t enemyy, uint16_t targetx, uint16_t targety){//finds the direction the enemy will take after the next tile
    uint8_t tmpdist = 0;
    uint8_t tmpdir;
    uint8_t tmp;
    if(enemydir == DIR_UP)
        enemyy-=1;
    else if(enemydir == DIR_LEFT)
        enemyx-=1;
    else if(enemydir == DIR_DOWN)
        enemyy+=1;
    else if(enemydir == DIR_RIGHT)
        enemyx+=1;
    if(enemyx == 0 || enemyx == 27)
        return enemydir;
    if(enemydir != DIR_DOWN && walls[enemyy-1][enemyx] < 3){
        tmpdist = distance(enemyx, enemyy-1, targetx, targety);
        tmpdir = DIR_UP;
    }
    if(enemydir != DIR_RIGHT && walls[enemyy][enemyx-1] < 3){
        if(!tmpdist || (tmp = distance(enemyx-1, enemyy, targetx, targety)) < tmpdist){
            tmpdist = tmp;
            tmpdir = DIR_LEFT;
        }
    }
    if(enemydir != DIR_UP && walls[enemyy+1][enemyx] < 3){
        if(!tmpdist || (tmp = distance(enemyx,enemyy+1,targetx,targety)) < tmpdist){
            tmpdist = tmp;
            tmpdir = DIR_DOWN;
        }
    }
    if(enemydir != DIR_LEFT && walls[enemyy][enemyx+1] < 3){
        if(!tmpdist || (tmp = distance(enemyx+1,enemyy,targetx,targety)) < tmpdist){
            tmpdist = tmp;
            tmpdir = DIR_RIGHT;
        }
    }
    return tmpdir;
}


void resetEnemies(uint8_t level){
    setuint16array(enemyx, (14*7)<<8,(12*7)<<8,(13*7+5)<<8,(15*7+3)<<8);
    setuint16array(enemyy, (11*7+3)<<8,(14*7+3)<<8,(14*7+3)<<8,(14*7+3)<<8);
    setuint8array(enemydir, DIR_LEFT,0,0,0);
    setuint8array(enemynextdir, DIR_LEFT,0,0,0);
    setuint8array(enemymode, 0,0,0,0);
    if(level == 1)
        return;
    if(level < 5)
        setuint16array(enemyvel, 218,206,202,192);
    else
        setuint16array(enemyvel, 244,238,238,230);
}

BOOL scattermode(uint8_t level){
    static BOOL alreadyscattered;
    if(elapsedtime(0) <= 7)
        return TRUE;
    else if((elapsedtime(0) > 7 + 20) && (elapsedtime(0) <= 7 + 20 + 7))
        return TRUE;
    else if((elapsedtime(0) > 7 + 20 + 7 + 20) && (elapsedtime(0) <= 7 + 20 + 7 + 20 + 5)){
        alreadyscattered = FALSE;
        return TRUE;
    }
    else if((level == 1) &&(elapsedtime(0) > 7 + 20 + 7 + 20 + 5 + 20) && (elapsedtime(0) <= 7 + 20 + 7 + 20 + 5 + 20 + 5))
        return TRUE;
    else if((level > 1) && (alreadyscattered == FALSE) && (elapsedtime(0) == 7 + 20 + 7 + 20 + 5 + 1035))
        return TRUE;
    return FALSE;
}

void moveEnemies(uint8_t level, uint16_t pacmanx, uint16_t pacmany){
    uint8_t scatterx[4] = {29,29,1,1};
    uint8_t scattery[4] = {1,26,1,26};
    static uint8_t enemylock[4] = {0,0,0,0};
    int i;

    if(elapsedtime(0) == STARTDELAY_PINKY){
        resetEnemy(2);
    }
    if(elapsedtime(0) == STARTDELAY_INKEY){
        resetEnemy(1);
    }
    if(elapsedtime(0) == STARTDELAY_CLYDE){
        resetEnemy(3);
    }

    for(i=0;i<4;i++){
        switch(enemydir[i]){
            case DIR_UP:
                enemyy[i] -= enemyvel[i];
            break;
            case DIR_LEFT:
                enemyx[i] -= enemyvel[i];
            break;
            case DIR_DOWN:
                enemyy[i] += enemyvel[i];
            break;
            case DIR_RIGHT:
                enemyx[i] += enemyvel[i];
            break;
        }
    }

    for(i=0;i<4;i++){
        if( !(((enemyx[i] >> 8) - 3) % 7) && enemylock[i] != ((enemyx[i]>>8)/7) + ((enemyy[i]>>8)/7) && !(((enemyy[i] >> 8) - 3) % 7) ){
            enemydir[i] = enemynextdir[i];
            if(!scattermode(level))
                enemynextdir[i] = newdir(enemydir[i], (enemyx[i]>>8)/7, (enemyy[i]>>8)/7, (pacmanx>>8)/7, (pacmany>>8)/7);
            else
                enemynextdir[i] = newdir(enemydir[i], (enemyx[i]>>8)/7, (enemyy[i]>>8)/7, scatterx[i], scattery[i]);
            enemylock[i] = ((enemyx[i]>>8)/7) + ((enemyy[i]>>8)/7);//it mustn't change its direction while still at the same place
        }
    }

    for(i=0;i<4;i++){
        if(enemyx[i]>>8 <= 4)
            enemyx[i] = (28*7 - 6)<<8;
        else if(enemyx[i]>>8 >= 28*7 - 5)
            enemyx[i] = 5<<8;
    }
}
