#include <os.h>

#include "level.h"
#include "pacmansprites.h"
#include "enemysprites.h"

#include "gfx.h"
#include "enemies.h"
#include "utils.h"
#include "timer.h"


//all coordinates of pacman and enemies are relative to the upper right corner of the board, i. e. from 0 to levelwidth*tilewidth - 1
//to do:    - start & death screens
//          - cornering
//          - look at the AI code again / fix it
//          - level "symbols" (strawberry, cherry, galaxian,...)
//          - boni
//optional: ~ geschwindigkeit erhhen: timer schneller
//          ~ bessere b/w: v.a. frightmode
//          ~ anzeige wie viele punkte beim bonus/geist fressen
//          ~ beim fressen langsamer & gegner im tunnel langsamer & gegner im frightmode langsamer
//          ~ eaten enemies wander around as eyes & return

/*changes since contest:
    - fixed memoryleak
    - fixed bug with parts of the levelbuf being cleared (happened because more was copied than was there)
    - made enemies work at any speed
    - enemies are reset when pacman dies
    - mehr aufgesplittet in functions und verschiedene dateien, dafr 5% - 15% weniger fps
    - cleaned up returns: buffers weren't freed upon dying & if one buffer couldn't be allocated, the others weren't freed
    - control framerate with a timer
    - better palette order for b/w -> playable on b/w calcs
    - better sprites for frightened enemies
    - 200, 400, 800, 1600 points for eating first, second, third, fourth enemy with one powerpill
    - scatter mode
*/


void drawEnemies(uint8_t* scrbuf, uint16_t* enemyx, uint16_t* enemyy, uint8_t* enemymode, uint8_t frighttime){
    uint8_t i,x,y,color;
    uint8_t (*sprite)[6];
    for(i=0;i<4;i++){
        if(enemymode[i] < 2)
            sprite = enemies[i];
        else if(frightenedtimeelapsed(0) + 2 < frighttime)
            sprite = frightenedenemy;
        else
            sprite = halffrightenedenemy;
        for(y=0;y<11;y++){
            for(x=0;x<11;x++){
                color = (x%2) ? sprite[y][x/2]&0x0F : sprite[y][x/2]>>4;
                if(color)
                    setPixelBuf(scrbuf, LEVEL_X_OFFSET+(enemyx[i]>>8)+x - (11/2), LEVEL_Y_OFFSET+(enemyy[i]>>8)+y - (11/2), color);
            }
        }
    }
}

void drawPacman(uint8_t* scrbuf, uint8_t pacmandir, uint16_t animx, uint16_t animy){
    static uint8_t frame = 0;
    uint8_t color;
    uint8_t x,y;
    uint8_t (*sprite)[6];

    if(pacmandir == 0)
        sprite = pacman;
    else if(pacmandir == DIR_RIGHT)
        sprite = pacman_right[frame/2];
    else if(pacmandir == DIR_LEFT)
        sprite = pacman_left[frame/2];
    else if(pacmandir == DIR_DOWN)
        sprite = pacman_down[frame/2];
    else if(pacmandir == DIR_UP)
        sprite = pacman_up[frame/2];
    else
        return;

    for(y=0;y<11;y++){
        for(x=0;x<11;x++){
            color = (x%2) ? sprite[y][x/2]&0x0F : sprite[y][x/2]>>4;
            if(color)
                setPixelBuf(scrbuf, LEVEL_X_OFFSET+animx+x - (11/2), LEVEL_Y_OFFSET+animy+y - (11 / 2), color);
        }
    }

    frame++;
    if(frame > 3)
        frame = 0;
}

void drawWalls(uint8_t* levelbuf){
    uint8_t x,y;
    for(y=0;y<31;y++){
        for(x=0;x<28;x++){
            if(walls[y][x] == 1 || walls[y][x] == 2)
                putSprite(levelbuf, walls[y][x], x*7, y*7, 0x2);
            else if(walls[y][x] == 25)
                putSprite(levelbuf, walls[y][x], x*7, y*7, 0x6);
            else
                putSprite(levelbuf, walls[y][x], x*7, y*7, 0x4);
        }
    }
}

int main(){
    uint8_t* scrbuf = (uint8_t*)malloc(SCREEN_BYTES_SIZE);
    uint8_t* levelbuf = (uint8_t*)malloc(31*7*28*7);
    uint8_t* levelresetbuf = (uint8_t*)malloc(31*28);
    memset(scrbuf,0xFF,SCREEN_BYTES_SIZE);
    memcpy(levelresetbuf,&walls[0][0],31*28);
    if(scrbuf == NULL || levelbuf == NULL || levelresetbuf == NULL){
        if(scrbuf != NULL)
            free(scrbuf);
        if(levelbuf != NULL)
            free(levelbuf);
        if(levelresetbuf != NULL)
            free(levelresetbuf);
        return 1;
    }

    if(is_cx)
        initPalette();
    initTimer();

    int i;

    uint16_t pacmanx;//pacmanx = x-coordinate of pacman's center point
    uint16_t pacmany;
    uint16_t pacmanvel;
    uint8_t pacmandir;
    uint8_t pacmannextdir;

    uint8_t* enemymode = getEnemymode();

    uint8_t eatendots;
    uint32_t score = 0;
    uint8_t lifes = 3;
    uint8_t level = 1;
    uint8_t frighttime = 6;

    uint16_t* enemyx = getEnemyx();
    uint16_t* enemyy = getEnemyy();

    uint8_t howmanyeaten=0;

    BOOL dead = FALSE;

    //drawTest(scrbuf);

    int fps = 0;
    uint32_t oldtime = 0;


    while(!isKeyPressed(KEY_NSPIRE_ESC) && !on_key_pressed() && !dead){
        pacmanx = (14*7) << 8;//pacmanx = x-coordinate of pacman's center point
        pacmany = (23*7 + 3) << 8;
        pacmandir = DIR_LEFT;
        pacmannextdir = pacmandir;

        if(level == 1)
            pacmanvel = 204;
        else if(level < 5)
            pacmanvel = 230;
        else if(level < 21)
            pacmanvel = 256;
        else
            pacmanvel = 230;

        eatendots = 0;
        frighttime = getFrighttime(level);
        resetEnemies(level);

        memcpy(walls, levelresetbuf, 31*28);
        drawWalls(levelbuf);
        drawLifes(scrbuf,lifes);
        elapsedtime(*(uint32_t*)0x90090000);

        while(!isKeyPressed(KEY_NSPIRE_ESC) && !on_key_pressed() && !dead){
            while(!timer())
                //idle();

            if(oldtime != *(uint32_t*)0x90090000){
                oldtime = *(uint32_t*)0x90090000;
                printscr(scrbuf,fps);
                printscr(scrbuf,scattermode(level));
                fps = 0;
            }
            ++fps;

            copyLevelbuf(scrbuf, levelbuf);
            drawPacman(scrbuf, pacmandir, pacmanx>>8, pacmany>>8);
            drawEnemies(scrbuf,enemyx,enemyy,enemymode,frighttime);
            drawScore(scrbuf,score);
            memcpy(*(uint8_t**)SCREEN_POINTER_ADDRESS,scrbuf,SCREEN_BYTES_SIZE);

            moveEnemies(level, pacmanx, pacmany);


            if(isKeyPressed(KEY_NSPIRE_UP))
                pacmannextdir = DIR_UP;
            else if(isKeyPressed(KEY_NSPIRE_LEFT))
                pacmannextdir = DIR_LEFT;
            else if(isKeyPressed(KEY_NSPIRE_RIGHT))
                pacmannextdir = DIR_RIGHT;
            else if(isKeyPressed(KEY_NSPIRE_DOWN))
                pacmannextdir = DIR_DOWN;

            if(!(((pacmanx >> 8) - 3) % 7) && !(((pacmany >> 8) - 3) % 7)){
                if(walls[(pacmany >> 8) / 7][(pacmanx >> 8) / 7] == 1){
                    walls[(pacmany >> 8) / 7][(pacmanx >> 8) / 7] = 0;
                    eatendots++;
                    score += 10;
                    drawWalls(levelbuf);
                }
                else if(walls[(pacmany >> 8) / 7][(pacmanx >> 8) / 7] == 2){
                    walls[(pacmany >> 8) / 7][(pacmanx >> 8) / 7] = 0;
                    eatendots++;
                    score += 50;
                    drawWalls(levelbuf);
                    startFrightmode();
                    howmanyeaten = 0;
                    frightenedtimeelapsed(*(uint32_t*)0x90090000);
                }

                if(eatendots == 244){
                    level++;
                    break;
                }

                switch(pacmandir){
                    case DIR_UP:
                        if(walls[((pacmany >> 8) / 7) - 1][(pacmanx >> 8) / 7] > 2)
                            pacmandir = 0;
                    break;
                    case DIR_LEFT:
                        if(walls[(pacmany >> 8) / 7][((pacmanx >> 8) / 7) - 1] > 2)
                            pacmandir = 0;
                    break;
                    case DIR_DOWN:
                        if(walls[((pacmany >> 8) / 7) + 1][(pacmanx >> 8) / 7] > 2)
                            pacmandir = 0;
                    break;
                    case DIR_RIGHT:
                        if(walls[(pacmany >> 8) / 7][((pacmanx >> 8) / 7) + 1] > 2)
                            pacmandir = 0;
                    break;
                }
            }

            if(pacmandir == DIR_UP)
                pacmany -= pacmanvel;
            if(pacmandir == DIR_LEFT)
                pacmanx -= pacmanvel;
            if(pacmandir == DIR_DOWN)
                pacmany += pacmanvel;
            if(pacmandir == DIR_RIGHT)
                pacmanx += pacmanvel;

            if( ((pacmandir | pacmannextdir) == (DIR_DOWN | DIR_UP)) || ((pacmandir | pacmannextdir) == (DIR_LEFT | DIR_RIGHT)))//if opposite directions change direction
                pacmandir = pacmannextdir;
            else if(!(((pacmanx >> 8) - 3) % 7) && !(((pacmany >> 8) - 3) % 7)){
                switch(pacmannextdir){
                    case DIR_UP:
                        if(walls[((pacmany >> 8) / 7) - 1][(pacmanx >> 8) / 7] < 3)
                            pacmandir = pacmannextdir;
                    break;
                    case DIR_LEFT:
                        if(walls[(pacmany >> 8) / 7][((pacmanx >> 8) / 7) - 1] < 3)
                            pacmandir = pacmannextdir;
                    break;
                    case DIR_DOWN:
                        if(walls[((pacmany >> 8) / 7) + 1][(pacmanx >> 8) / 7] < 3)
                            pacmandir = pacmannextdir;
                    break;
                    case DIR_RIGHT:
                        if(walls[(pacmany >> 8) / 7][((pacmanx >> 8) / 7) + 1] < 3)
                            pacmandir = pacmannextdir;
                    break;
                }

                //else{//if 90 degree change go diagonally (or if same direction as before do not change anything)
                //    pacmandir |= pacmannextdir;
                //}
            }

            if(pacmanx>>8 <= 4)
                pacmanx = (28*7-6)<<8;
            else if(pacmanx>>8 >= 28*7 - 5)
                pacmanx = 5<<8;

            for(i=0;i<4;i++){
                if(((pacmanx>>8)/7 == (enemyx[i]>>8)/7) && ((pacmany>>8)/7 == (enemyy[i]>>8)/7)){
                    if(enemymode[i] < 2){
                        if(lifes == 0){
                            dead = TRUE;
                            break;
                        }
                        lifes--;
                        pacmanx = (14*7) << 8;
                        pacmany = (23*7 + 3) << 8;
                        pacmandir = DIR_LEFT;
                        drawLifes(scrbuf, lifes);
                        resetEnemies(level);
                        sleep(2000);
                        elapsedtime(*(uint32_t*)0x90090000);
                    }
                    else{
                        resetEnemy(i);
                        resetEnemymode(i);
                        score += (2<<howmanyeaten)*100;
                        howmanyeaten++;
                    }
                }
            }
            if(frightenedtimeelapsed(0) > frighttime)
                endFrightmode();
        }
    }
    free(scrbuf);
    free(levelbuf);
    free(levelresetbuf);
    restoreTimer();
    lcd_incolor();
    return 0;
}
