#include <os.h>
#include "timer.h"

#define CX_TIMER_CONTROLREG (*(volatile uint8_t*)0x900D0028)
#define BW_TIMER_CONTROLREG (*(volatile uint16_t*)0x900D0010)
#define CX_TIMER_VALUE (*(volatile uint32_t*)0x900D0024)
#define BW_TIMER_VALUE (*(volatile uint32_t*)0x900D000C)

static uint16_t restorecontrolreg;

void initTimer(){
    if(is_cx){
        restorecontrolreg = CX_TIMER_CONTROLREG;
        CX_TIMER_CONTROLREG = CX_TIMER_CONTROLREG & ~0x80;//disable timer
        CX_TIMER_CONTROLREG = (CX_TIMER_CONTROLREG & ~0b00101100) | 0b1000;//prescaler (divider) = 256 && interrupts disabled
        CX_TIMER_CONTROLREG = CX_TIMER_CONTROLREG | 0x80;//re-enable timer
    }
    else{
        restorecontrolreg = BW_TIMER_CONTROLREG;
        BW_TIMER_CONTROLREG = 510;//divider = 510
    }
}

void restoreTimer(){
    if(is_cx){
        CX_TIMER_CONTROLREG = CX_TIMER_CONTROLREG & ~0x80;//disable timer
        CX_TIMER_CONTROLREG = restorecontrolreg;
        CX_TIMER_CONTROLREG = CX_TIMER_CONTROLREG | 0x80;//re-enable timer
    }
    else
        BW_TIMER_CONTROLREG = restorecontrolreg;
}

int timer(){
    static uint32_t oldtimerval = 0;
    static BOOL pass = FALSE;
    if(is_cx && (oldtimerval != CX_TIMER_VALUE)){
        pass = !pass;
        oldtimerval = CX_TIMER_VALUE;
        if(!pass)
            return 1;
        return 0;
    }
    if(!is_cx && (oldtimerval != BW_TIMER_VALUE)){
        oldtimerval = BW_TIMER_VALUE;
        return 1;
    }
    return 0;
}
